"""The scripts module contains executable modules that provide simple utility
scripts showcasing the abilities of simple_3dviz and allowing for the quick
visualization of two-dimensional functions (`simple_3dviz.scripts.func_viewer`)
and meshes (`simple_3dviz.scripts.mesh_viewer`)."""
